#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 15 20:29:09 2017

@author: tommy
"""

import random
import functools
import math
import statistics
import operator


def chunker(iterable, size=5, fill=''):
    """
    """

    for i in range(0, len(iterable) // size + 1):
        to_yield = iterable[i * size: (i + 1) * size]
        if len(to_yield) == 0:
            break
        if len(to_yield) < size:
            yield to_yield + [fill] * (size - len(to_yield))
        else:
            yield to_yield

def prioritized_not_None(*args):
    """Return the first argument which is not None.

    Parameters
    ----------
    *args
        Arguments to the function.

    Returns
    -------
    argument
        The first argument which is not None.


    Examples
    -------
    >>> prioritized_not_None(None, None, 2, 5)
    2
    """
    for arg in list(args):
        if arg is None:
            continue
        return arg
    return None


def round_to_nearest(number, nearest = 1):
    """Round 'number' to the nearest multiple of 'nearest'.

    Parameters
    ----------
    number
        A real number to round.
    nearest
        Number to round to closes multiple of.

    Returns
    -------
    rounded
        A rounded number.


    Examples
    -------
    >>> round_to_nearest(6.8, nearest = 2.5)
    7.5
    """
    return nearest * round(number / nearest)


def all_equal(iterable):
    """Checks whether all items in an iterable are equal.

    Parameters
    ----------
    iterable
        An iterable, e.g. a string og a list.

    Returns
    -------
    boolean
        True or False.
    
    Examples
    -------
    >>> all_equal([2, 2, 2])
    True
    >>> all_equal([1, 2, 3])
    False
    """
    first = iterable[0]
    return all([first == i for i in iterable[1:]])


def min_between(min_reps = 3, max_reps = 8, percentile = 0.33):
    """
    Function to decide the minimum number of reps to perform
    given `min_reps` and `max_rep`.
    
    Examples
    -------
    >>> min_between(min_reps = 3, max_reps = 8, percentile = 0.33)
    (3, 5)
    """
    higher_limit = min_reps + (max_reps - min_reps) * percentile
    return min_reps, math.ceil(higher_limit)


def generate_reps(min_reps = 3, max_reps = 8, total = 25, existing = None):
    """Generate 'total' repetitions between 'min_reps' and 'max_reps',
    if existing is given (not None), then some repetitions are already
    drawn.

    Parameters
    ----------
    min_reps
        Lower limit for the repetitions dawing, e.g. 3.
        
    max_reps
        Upper limit for the repetitions dawing, e.g. 8.
        
    total
        The total number of repetitions to return.
        
    existing
        A list of prior reps drawn.

    Returns
    -------
    reps_list
        A list of repetitions between 'min_reps' and 'max_reps', summing to 'total'
        or a number close to it.


    Examples
    -------
    >>> total = 20
    >>> reps_list = generate_reps(3, 8, total)
    >>> sum(reps_list) == total
    True
    """
    
    # Attempt to generate desired number of repetitions 10 times
    attempt = _generate_reps(min_reps, max_reps, total, existing)
    for i in range(10):
        attempt = _generate_reps(min_reps, max_reps, total, existing)
        if sum(attempt) == total:
            return list(map(int, attempt))

    # Convert to integer so there are no int/float mixups
    return list(map(int, attempt))
    

def _generate_reps(min_reps = 3, max_reps = 8, total = 25, existing = None):
    """
    Helper function to generate repetitions.
    """
    
    # Make a copy so as not to change the input variable
    if existing is None:
        existing = []
    else:
        existing = existing.copy()
        
    # While the sum is low, keep adding
    while sum(existing) + max_reps <= total:
        existing.append(random.randint(min_reps, max_reps))
    
    # If the missing number is between limits, add it to get the total
    missing = total - sum(existing)
    if min_reps <=  missing <= max_reps:
        existing.append(missing)

    # Sort the result
    existing.sort(reverse = True)

    return existing

def spread(iterable):
    """Returns the maximal spread of a sorted list of numbers.

    Parameters
    ----------
    iterable
        A list of numbers.

    Returns
    -------
    max_diff
        The maximal difference when the iterable is sorted.


    Examples
    -------
    >>> spread([1, 11, 13, 15])
    10
    
    >>> spread([1, 15, 11, 13])
    10
    """
    if len(iterable) == 1:
        return 0
    
    iterable = iterable.copy()
    iterable.sort()
    
    max_diff = max(abs(i-j) for (i, j) in zip(iterable[1:], iterable[:-1]))

    return max_diff


    
if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)
    
    
    
    
    
    
    
    
    
    
    
    
    
    
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 15 21:01:51 2017

@author: tommy
"""

from streprogen.modeling import (RepellentGenerator, 
                              reps_to_intensity,
                              reps_to_intensity_tight, 
                              reps_to_intensity_relaxed,
                              progression_linear, 
                              progression_sinusoidal)

from streprogen.exercises import (StaticExercise, DynamicExercise)

from streprogen.day import (Day)

from streprogen.program import (Program)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 15 21:18:22 2017

@author: tommy
"""

import functools
import operator
import statistics
import warnings
from jinja2 import Environment, PackageLoader, select_autoescape

from streprogen.utils import (round_to_nearest, all_equal, min_between,
                   spread, prioritized_not_None, 
                   generate_reps, chunker)

from streprogen.modeling import (RepellentGenerator, reps_to_intensity_tight, 
                   reps_to_intensity, progression_sinusoidal)

from streprogen.exercises import (StaticExercise, DynamicExercise)

from streprogen.day import Day

class ProgramError(Exception):
    pass

class Program(object):
    """
    The program class is a container for days and exercises,
    along with the methods and functions used to create training programs.
    """
    
    REP_SET_SEP = ' x '
    
    def __init__(self, 
                 name = 'Untitled', 
                 duration = 8, 
                 reps_per_exercise = 25, 
                 reps_scalers = None, 
                 avg_intensity = 75,
                 intensity_scalers = None,
                 progression_func = None,
                 reps_to_intensity_func = None, 
                 units = 'kg',
                 round_to = 2.5,
                 min_reps_consistency = None,
                 minimum_percentile = 0.2,
                 go_to_min = False,
                 times_to_render = 100,
                 verbose = False):
        """Initialize a new program.
    
        Parameters
        ----------
        name
            The name of the training program, e.g. 'Tommy_Spring_2017'.
            
        duration
            The duration of the training program in weeks, typically 4-12 weeks.
            
        reps_per_exercise
            Number of baseline repetitions per dynamic exercise, typically 20-35 repetitions.
            This can be overridden in a specific dynamic exercise by passing
            the 'reps_per_exercise' argument when constructing the dynamic exercise.
            
        reps_scalers
            A list of scaling factors for the repetitions, of the same length as
            the 'duration' variable, e.g. [1, 1.1, 0.9, 1.2, ...].
            
        avg_intensity
            Baseline average intensity for dynamic exercises, e.g. 75. The average
            intensity is the percentage of maximum that the average set is
            done with, weighted by the number of reptitions. Mathematically,
            avg_intensity = (sum_{sets} intensity * reps) / (sum_{sets} reps).
            This can be overridden in a specific dynamic exercise by passing
            the 'avg_intensity' argument when constructing the dynamic exercise.
            
        intensity_scalers
            A list of scaling factors for the average intensity, of the same length
            as 'duration' variable, e.g. [1, 1.1, 0.9, 1.2, ...].
            
        progression_func
            A function mapping from a week to the current 1RM, i.e.
            it should have a signature like:
                progression_func(week, start_weight, end_weight, start_week, end_week)
            
        reps_to_intensity_func
            A function mapping from a repetition to an intensity, i.e.
            it should map [1, 20] to [100, 40] and be decreasing.
            
        units
            The weight units, e.g. 'kg' or 'lbs'.
            
        round_to
            Round the weights of closes multiple of this number, e.g. 2.5 or 5.
            
        min_reps_consistency
            Override the 'exercise',  'weekly' or 'daily' modes of minimum repetitions.
            Overriding is only possible in one direction.
            
        minimum_percentile
            Percentile of the repeition range to draw the 'min_rep' variable
            from. If 'minimum_percentile' is 0.5 for and a dynamic
            exercise has min_reps = 4 and max_reps = 7,
            the minimum repetition always be drawn from [4, 5].
            If 'minimum_percentile' is 1, it will be drawn from [4,5,6,7].
            The 'min_rep' is what the program is minimally allowed to go to
            for that specific week/day/exercise.
            
        go_to_min
            Every week, day or for every exercise, a 'min_rep' value is 
            drawn from the percentile given by 'minimum_percentile'.
            Setting 'go_to_min' to True forces the program to include
            this repetition value in every repstring.
            
        times_to_render
            The number of times to render the dynamic exercises before
            picking the best generated values, default is 50.
            
        verbose
            If True, information is printed when the program is rendered.
            
    
        Returns
        -------
        Program
            A Program instance.
    
    
        Examples
        -------
        >>> program = Program('My training program')
        >>> program._rendered
        False
        """
        self.name = name
        self.duration = duration 
        self.reps_per_exercise = reps_per_exercise
        self.avg_intensity = avg_intensity
        
        # Set default value for rep_scalers if None
        if reps_scalers is None:
            # Draw self-repellent numbers from domain
            domain = [0.8, 1, 1.2]
            generator = RepellentGenerator(domain)
            self.reps_scalers = list(generator.yield_from_domain(self.duration))
        else:
            if len(reps_scalers) != self.duration:
                raise ProgramError('Length of `reps_scalers` must match program duration.')
            self.reps_scalers = reps_scalers
            
        # Set default value for intensity_scalers if None
        if intensity_scalers is None:
            # Draw self-repellent numbers from domain
            domain = [0.95, 1, 1.05]
            generator = RepellentGenerator(domain)
            self.intensity_scalers = list(generator.yield_from_domain(self.duration))
        else:
            if len(intensity_scalers) != self.duration:
                raise ProgramError('Length of `intensity_scalers` must match program duration.')
            self.intensity_scalers = intensity_scalers
            
        self.progression_func = prioritized_not_None(progression_func,
                                                     progression_sinusoidal)
        
        self.reps_to_intensity_func = prioritized_not_None(reps_to_intensity_func,
                                                           reps_to_intensity)
        self.units = units
        self.round = functools.partial(round_to_nearest, nearest = round_to)
        self.days = []
        self.min_reps_consistency = min_reps_consistency
        self.minimum_percentile = minimum_percentile
        self.go_to_min = go_to_min
        self.times_to_render = times_to_render
        self.verbose = verbose
        self._rendered = False
        
        
        
        
    def _validate(self):
        """
        The purpose of this method is to verify that the user has set sensible
        values for the training program before rendering. The user will still
        be able to render, but error messages will be printed. This method:
            
            * Validates that the average intensity is in the range [65, 85].
            * Validates that the number of repetitions is in the range [15, 45].
            * Validates that 'reps_to_intensity_func' maps to [0, 100].
            * Validates that 'reps_to_intensity_func' is a decreasing function.
            * Validates that the exercises do not grow more than 2.5% per week.
            
        Apart from these sanity checks, the user is on his own.
        """
        # Validate the intensity
        if max([s * self.avg_intensity for s in self.intensity_scalers]) > 85:
            warnings.warn('\nWARNING: Average intensity is > 85.')
            
        if min([s * self.avg_intensity for s in self.intensity_scalers]) < 65:
            warnings.warn('\nWARNING: Average intensity is < 65.')
            
        # Validate the repetitions
        if max([s * self.reps_per_exercise for s in self.reps_scalers]) > 45:
            warnings.warn('\nWARNING: Number of repetitions > 45.')
            
        if min([s * self.reps_per_exercise for s in self.reps_scalers]) < 15:
            warnings.warn('\nWARNING: Number of repetitions < 15.')
            
        # Validate the 'reps_to_intensity_func'
        for x1, x2 in zip(range(1, 20), range(2, 21)):
            y1 = self.reps_to_intensity_func(x1)
            y2 = self.reps_to_intensity_func(x2)
            if y1 < y2:
                warnings.warn("\n'reps_to_intensity_func' is not decreasing.")
        
        if any(self.reps_to_intensity_func(x) > 100 for x in range(1, 20)):
            warnings.warn("\n'reps_to_intensity_func' maps to > 100.")
            
        if any(self.reps_to_intensity_func(x) < 0 for x in range(1, 20)):
            warnings.warn("\n'reps_to_intensity_func' maps to < 0.")
            
            
        # Validate the exercises
        for day in self.days:
            for dynamic_ex in day.dynamic_exercises:
                start, end = dynamic_ex.start_weight, dynamic_ex.end_weight
                percentage_growth = (end/start)**(1/self.duration)
                percentage_growth = dynamic_ex.weekly_growth(self.duration)
                if percentage_growth > 2.5:
                    msg = '\n"{}" grows with {}% each week.'.format(
                            dynamic_ex.name, percentage_growth)
                    warnings.warn(msg)


    def add_days(self, *days):
        """Add several days to the prorgram.
    
        Parameters
        ----------
        list_of_days
            A list of days, e.g. [day1, day2, ...]
    
    
        Examples
        -------
        >>> program = Program('My training program')
        >>> day1, day2 = Day(), Day()
        >>> program.add_days([day1, day2])
        """
        for day in list(days):
            self.days.append(day)
            
        
    @staticmethod
    def repstring_penalty(reps, intensities, desired_reps, desired_intensity,
                           minimum_rep):
        """Penalty function which calculates how "bad" a set of
        reps and intensities is, compared to the desired repetitions,
        the desired intensity level and the minimum repetitions.
        Advanced users may substitute this function for their own version.
    
        Parameters
        ----------
        reps
            A list of repetitions (sorted), e.g. [8, 6, 5, 2].
        intensities
            A list of intensities corresponding to the repetitions,
            e.g. [64.7, 72.3, 76.25, 88.7].
        desired_reps
            Desired number of repetitions in total, e.g. 25.
        desired_intensity
            The desired average intensity, e.g. 75.
        minimum_rep
            The minimum repetition which is allowed, e.g. 2.
    
    
        Returns
        -------
        penalty
            A penalty, a positive real number.
    
    
        Examples
        -------
        >>> desired_reps = 25
        >>> desired_intensity = 75
        >>> minimum_rep = 1
        >>> high = Program().repstring_penalty([8, 8, 8], [60, 60, 60], 
        ...                              desired_reps, desired_intensity, 
        ...                              minimum_rep)
        >>> low = Program().repstring_penalty([8, 6, 5, 4, 2], [64, 72, 75, 80, 88], 
        ...                              desired_reps, desired_intensity, 
        ...                              minimum_rep)
        >>> high > low
        True
        """
        # Punish when the mean intensity is far from the desired one
        desired = desired_intensity
        error1 = abs(statistics.mean(intensities) - desired)
        
        # Punish when the repeitions are far from the desired amount
        error2 = abs(sum(reps) - desired_reps)
        
        # Punish when the spread of repetitions is large
        error3 = spread(reps)
        
        # Punish deviation from the minimum reptition
        error4 = abs(min(reps) - minimum_rep)

        # Take a linear combination and return.
        errors = [2 * error1, 0.5 * error2, 2.5 * error3, 0.5 * error4]

        #print(*[round(err, 4) for err in errors], sep = '\t')
        return sum(errors)
    
        
    def _render_dynamic(self, dynamic_exercise, minimum_rep, 
                                 desired_reps, desired_intensity):
        """
        Render a single dynamic exercise.
        This is done for each exercise every week.
        """
        
        # --------------------------------
        # Generate possible repstring and calculate penalties
        # --------------------------------
        generated_repstrings = []
        for k in range(self.times_to_render):
            if self.go_to_min:
                reps = generate_reps(minimum_rep,
                                 dynamic_exercise.max_reps,
                                 desired_reps - minimum_rep,
                                 [minimum_rep])
            else:
                reps = generate_reps(minimum_rep,
                                 dynamic_exercise.max_reps,
                                 desired_reps)
            
            intensities = list(map(self.reps_to_intensity_func, reps))

            penalty = self.repstring_penalty(reps,
                                          intensities,
                                          desired_reps,
                                          desired_intensity,
                                          minimum_rep)
            
            generated_repstrings.append((penalty, reps, intensities))
            
        # --------------------------------
        # Find the best generated repstring and verify it
        # --------------------------------
        best_repstring = min(generated_repstrings)
        (penalty, reps, intensities) = best_repstring
        
        # Perform a sanity check:
        # If repetitions are too high, a low average intensity cannot be attained
        if desired_intensity > self.reps_to_intensity_func(minimum_rep):
            msg = """
WARNING: The exercise '{}' is restricted to repetitions in the range [{}, {}],
but the desired average intensity for this week is {}. Reaching this intensity
is not attainable since it corresponds to repetitions lower than {}.
SOLUTION: Either (1) allow lower repetitions, (2) change the desired intensity
or (3) ignore this message. The software will do it's best to remedy this.
""".format(dynamic_exercise.name, dynamic_exercise.max_reps, minimum_rep,
desired_intensity, minimum_rep) 
            warnings.warn(msg)

        # Perform a sanity check:
        # If repetitions are too low, a high average intensity cannot be attained
        if desired_intensity < self.reps_to_intensity_func(dynamic_exercise.max_reps):
            msg = """
WARNING: The exercise '{}' is restricted to repetitions in the range [{}, {}],
but the desired average intensity for this week is {}. Reaching this intensity
is not attainable since it corresponds to repetitions higher than {}.
SOLUTION: Either (1) allow higher repetitions, (2) change the desired intensity
or (3) ignore this message. The software will do it's best to remedy this.
""".format(dynamic_exercise.name, dynamic_exercise.max_reps, minimum_rep,
desired_intensity, dynamic_exercise.max_reps) 
            warnings.warn(msg)
            
            
        
        if self.verbose:
            print('Found optimal repstring for "{}".'.format(dynamic_exercise.name))
            print(' Reps   (desired/actual):', int(desired_reps), int(sum(reps)), sep = '\t')
            print(' MI     (desired/actual):', desired_intensity, 
              round(statistics.mean(intensities), 3), sep = '\t')
            print(' minrep (limit/actual): ', minimum_rep, 
              min(reps), sep = '\t')
            print(' largest jump:          ', spread(reps), sep = '\t')

            print(best_repstring[1],best_repstring[0],  sep = '\t')
            print()
        

        return {'reps' : reps, 'intensities' : intensities}

        

        
    def _initialize_render_dictionary(self):
        """Initialize a dictionary for rendered values.
    
    
        Examples
        -------
        >>> program = Program('My training program')
        >>> program._initialize_render_dictionary()
        >>> program._rendered is False
        False
        """

        self._rendered = dict()
        
        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            self._rendered[week] = dict()
            
            # Iterate over all days
            for day in self.days:
                self._rendered[week][day] = dict()
                
                # Iterate over all main exercises
                for dynamic_ex in day.dynamic_exercises:
                    self._rendered[week][day][dynamic_ex] = dict()
                    
    def _set_min_reps(self):        
        """Populate the _rendered dictionary with entries corresponding
        to the minimum number of reps to go to for each exercise
        and for the entire duration.
    
    
        Examples
        -------
        >>> program = Program('My training program', duration = 2)
        >>> bench_press = DynamicExercise('Bench', 100, 120)
        >>> day = Day(exercises = [bench_press])
        >>> program.add_days(day)
        >>> program._initialize_render_dictionary()
        >>> program._autoset_min_reps_consistency()
        >>> program._set_min_reps()
        >>> for week in range(1, program.duration + 1):
        ...     for day in program.days:
        ...         for dynamic_ex in day.dynamic_exercises:
        ...             print(program._rendered[week][day][dynamic_ex]['minimum'] > 0)
        True
        True
        """

        # --------------------------------
        # If the mode is weekly, set minimum reps on a weekly basis
        # --------------------------------
        if self._min_reps_consistency == 'weekly':
            
            # Set up generator. Only one is needed
            exercise = self.days[0].dynamic_exercises[0]
            low, high = min_between(exercise.min_reps, 
                                    exercise.max_reps, 
                                    self.minimum_percentile)
            generator = RepellentGenerator(list(range(low, high+1)))

            # Use generator to populate the dictionary with minimum values
            for week in range(1, self.duration + 1):
                minimum_rep_this_week = generator.generate_one()
                for day in self.days:
                    for dynamic_ex in day.dynamic_exercises:
                        self._rendered[week][day][dynamic_ex]['minimum'] = minimum_rep_this_week
                        
                                      
        # --------------------------------
        # If the mode is daily, set minimum reps on a daily basis
        # --------------------------------
        if self._min_reps_consistency == 'daily':
            
            # Set up generators. One is needed for each day
            generators = dict()
            for day in self.days:
                exercise = day.dynamic_exercises[0]
                low, high = min_between(exercise.min_reps, 
                                        exercise.max_reps, 
                                        self.minimum_percentile)
                generator = RepellentGenerator(list(range(low, high+1)))
                generators[day] = generator

            # Use generators to populate the dictionary with minimum values
            for week in range(1, self.duration + 1):
                for day in self.days:
                    minimum_rep_this_day = generators[day].generate_one()
                    for dynamic_ex in day.dynamic_exercises:
                        self._rendered[week][day][dynamic_ex]['minimum'] = minimum_rep_this_day

        # --------------------------------
        # If the mode is by exercise, set minimum reps on an exercise basis
        # --------------------------------           
        if self._min_reps_consistency == 'exercise':
            
            # Set up generators. One is needed for each exercise
            generators = dict()
            for day in self.days:
                for dynamic_ex in day.dynamic_exercises:
                    low, high = min_between(dynamic_ex.min_reps, 
                                            dynamic_ex.max_reps, 
                                            self.minimum_percentile)
                    generator = RepellentGenerator(list(range(low, high+1)))
                    generators[dynamic_ex] = generator
                              
            # Use generators to populate the dictionary with minimum values
            for week in range(1, self.duration + 1):
                for day in self.days:
                    for dynamic_ex in day.dynamic_exercises:
                        minimum_rep_this_ex = generators[dynamic_ex].generate_one()
                        self._rendered[week][day][dynamic_ex]['minimum'] = minimum_rep_this_ex
        
        
    def _yield_week_day_dynamic(self):
        """A helper function to reduce the number of nested loops.

        Yields
        -------
        (week, day, dynamic_ex)
            A tuple with information.

        """
        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            # Iterate over all days
            for day in self.days:
                # Iterate over all main exercises
                for dynamic_ex in day.dynamic_exercises:
                    yield (week, day, dynamic_ex)
                    
    def _yield_week_day(self, enumeration = False):
        """A helper function to reduce the number of nested loops.
        
        Parameters
        ----------
        enumeration
            Whether or not to wrap the days in enumerate().
            
    
        Yields
        -------
        (week, day_index, day) or (week, day)
            A tuple with information, depending on 'enumeration' parameter.

        """
        if enumeration:
            # Iterate over all weeks
            for week in range(1, self.duration + 1):
                # Iterate over all days
                for day_index, day in enumerate(self.days):
                    yield (week, day_index, day)
        else:
            # Iterate over all weeks
            for week in range(1, self.duration + 1):
                # Iterate over all days
                for day in self.days:
                    yield (week, day)

    def _yield_dynamic_exercises(self):
        """A helper function to reduce the number of nested loops.


        Yields
        -------
        (dynamic_ex)
            Yields the dynamic exercises in the program.

        """
        for day in self.days:
            for dynamic_ex in day.dynamic_exercises:
                yield dynamic_ex
                    
    def _yield_exercises(self):
        """A helper function to reduce the number of nested loops.
        

        Yields
        -------
        (dynamic_ex) or (static_ex)
            Yields the exercises in the program.

        """
        for day in self.days:
            for dynamic_ex in day.dynamic_exercises:
                yield dynamic_ex
            for static_ex in day.static_exercises:
                yield static_ex
        
    def render(self, validate_first = True):
        """Render the training program to perform the calculations.
        The program can be rendered several times to produce new
        information given the same input parameters.
    
        Parameters
        ----------
        validate_first
            Boolean that indicates whether or not to run a validation
            heurestic on the program before rendering. The validation
            will warn the user if inputs seem unreasonable.

        """
        
        # --------------------------------
        # Prepare for rendering the dynamic exercises
        # --------------------------------
        
        # Set the minimum consistency mode of the program
        if self.min_reps_consistency is None:
            self._autoset_min_reps_consistency()
        else:
            self._autoset_min_reps_consistency()
            
            # Make sure the user defined consistency mode is
            # never more broad than what is allowed by inputs
            if (self._min_reps_consistency == 'exercise' and
                self.min_reps_consistency != 'exercise'):
                    raise ProgramError("Error with 'min_reps_consistency'.")
                    
            if (self._min_reps_consistency == 'daily' and
                self.min_reps_consistency == 'weekly'):
                    raise ProgramError("Error with 'min_reps_consistency'.")

            self._min_reps_consistency = self.min_reps_consistency
        
        # Check the minimum consistency mode of the program in
        # relation to the user input
        
            
        # Initialize the structure of the _rendered dictionary
        self._initialize_render_dictionary()
        
        # Set the day names
        for i, day in enumerate(self.days):
            day.name = prioritized_not_None(day.name, 'Day {}'.format(i + 1))
        
        # Set the minimum reps per week in the render dictionary
        self._set_min_reps()
        
        # Validate the program
        if validate_first:
            self._validate()
            
        
        # --------------------------------
        # Render the dynamic exercises
        # --------------------------------
            
        for (week, day, dynamic_ex) in self._yield_week_day_dynamic():
            
                    
            # The minimum repeition to work up to
            min_rep = self._rendered[week][day][dynamic_ex]['minimum']

            # The reps to work up to
            total_reps = prioritized_not_None(dynamic_ex.reps, 
                                              self.reps_per_exercise)
            desired_reps = total_reps * self.reps_scalers[week - 1]
            self._rendered[week][day][dynamic_ex]['desired_reps'] = int(desired_reps)
            
            # The desired intensity
            intensity_unscaled = prioritized_not_None(dynamic_ex.avg_intensity,
                                                      self.avg_intensity)
            intensity_scale = self.intensity_scalers[week - 1]
            desired_intensity = intensity_unscaled * intensity_scale
            self._rendered[week][day][dynamic_ex]['desired_intensity'] = int(desired_intensity)
            
            # A dictionary returned with keys 'reps' and 'intensities'
            out = self._render_dynamic(dynamic_ex, 
                                       min_rep,
                                       desired_reps,
                                       desired_intensity)
            
            weight = self.progression_func(week, 
                                           dynamic_ex.start_weight, 
                                           dynamic_ex.end_weight,
                                           start_week = 1, 
                                           end_week = self.duration)

            def pretty_weight(weight, i, round_func):
                weight = round_func(weight*i/100)
                if weight % 1 == 0:
                    return int(weight)
                return weight

            round_func = prioritized_not_None(dynamic_ex.round, self.round)
            
            out['strings'] = [self.REP_SET_SEP.join(
                    [str(r), str(pretty_weight(weight, i, round_func)) + self.units])
                    for (i, r) in zip(out['intensities'], out['reps'])]
            
            # Update the dictionary
            self._rendered[week][day][dynamic_ex].update(out)
            #print(out['string'])
            #print(self._rendered[week][day][dynamic_ex])
                    
                    

    def to_html(self, verbosity = 0):
        """

        :param verbosity:
        :return:
        """


        env = Environment(
            loader=PackageLoader('streprogen', 'templates'),
            autoescape=select_autoescape(['html', 'xml']),
            trim_blocks = True,
            lstrip_blocks = True
        )

        template = env.get_template('program_template.html')
        env.globals.update(chunker=chunker, enumerate = enumerate)

        return template.render(program=self, table_width=5)


    def to_text2(self):
        env = Environment(
            loader=PackageLoader('streprogen', 'templates'),
            autoescape=select_autoescape(['html', 'xml']),
            trim_blocks = True,
            lstrip_blocks = True
        )

        template = env.get_template('program_template.txt')


        return template.render(program=self, table_width = 5)


    
    def to_text(self, verbosity = 0):
        """Return a string with the program information.
    
        Parameters
        ----------
        verbosity
            How much information to include about the program.
                0 - Include basic information only.  
                
                1 - Include some statistical information.
                
                2 - Include all statistical information.
    
        Returns
        -------
        program_string
            A string with program information.

        """
        
        # --------------------------------
        # Get information related to lengths
        # --------------------------------
        
        # Find the length of the longest exercise name
        l_ex_name = max(len(ex.name) for ex in self._yield_exercises())
        
        # If rendered, find the length of the longest '6 x 75kg'-type string
        l_ex_str = 0
        if self._rendered:
            for (week, day, dynamic_ex) in self._yield_week_day_dynamic():
                lengths = [len(s) for s in 
                           self._rendered[week][day][dynamic_ex]['strings']]
                l_ex_str = max(l_ex_str, max(lengths))
                        
        # --------------------------------
        # Header information
        # --------------------------------

        # Print the start of the header, some parameters from the setup
        out_str = '-' * 64 + '\n'
        out_str += 'Program: {} \n'.format(self.name)
        keys_to_print = ['duration', 'reps_per_exercise', 'avg_intensity', 
        'reps_scalers', 'intensity_scalers', 'units']
        out_str += '\nParameters\n'
        for key in keys_to_print:
            out_str += '  {}: {}\n'.format(key, self.__dict__[key])
        
        # --------------------------------
        # Day and exercise information
        # --------------------------------
        out_str += '-' * 64 + '\n'
        out_str += 'Days and exercises\n'
        for i, day in enumerate(self.days):
            out_str += '  ' + prioritized_not_None(day.name, 
            'Day {}'.format(i + 1)) + '\n'
            
            # Loop for dynamic exercises
            for dynamic_ex in day.dynamic_exercises:
                min_reps = str(dynamic_ex.min_reps)
                max_reps = str(dynamic_ex.max_reps)
                start_weight = str(dynamic_ex.start_weight).rjust(3)
                end_weight = str(dynamic_ex.end_weight).rjust(3)
                units = self.units
                num_reps = dynamic_ex.reps
                increase = dynamic_ex.weekly_growth(self.duration)
                ex_int = dynamic_ex.avg_intensity
                
                # Basic exercise information
                name = ' '*3 + dynamic_ex.name.ljust(l_ex_name + 2)
                weights = '{} -> {}'.format(start_weight + units, 
                                            end_weight + units)
                reps_ra = 'reps: [{}, {}]'.format(min_reps, max_reps)
                weekly_inc = 'weekly inc.: {}%'.format(increase)
                
                # Advaned exercise information
                reps = prioritized_not_None(num_reps, self.reps_per_exercise)
                reps = 'reps/ex: {}'.format(int(reps))
                inten = prioritized_not_None(ex_int, self.avg_intensity)
                inten = 'intensity: {}'.format(int(inten))
                
                include_data = [name, weights, reps_ra, weekly_inc]
                if verbosity > 0:
                    include_data = include_data[:-1] + [reps, inten] + include_data[-1:]

                out_str += ''.join(s.ljust(16) for s in include_data)   
                out_str += '\n'
                
            # Loop for static exercises
            for static_ex in day.static_exercises:
                name = ' '*3 + static_ex.name.ljust(l_ex_name + 2)
                scheme = static_ex.sets_reps
                out_str += ''.join(s.ljust(16) for s in [name, scheme])
                out_str += '\n'
        
            
        # --------------------------------
        # The set/rep schemes for all weeks
        # --------------------------------
        if not self._rendered:
            return out_str + '-' * 64
        out_str += '-' * 64 + '\n'
        # Iterate over all weeks
        for week in range(1, self.duration + 1):
            week_str = 'Week {}'.format(week)
            if verbosity > 0:
                week_str += ' '*4 + '[Avg. intensity: {}, Reps: {}]'.format(
                        int(self.avg_intensity * self.intensity_scalers[week-1]),
                        int(self.reps_per_exercise * self.reps_scalers[week-1]))
            out_str += week_str + '\n'
            
            # Iterate over all days
            for i, day in enumerate(self.days):
                out_str += ' ' + day.name + '\n'
                
                # Iterate over all main exercises
                for dynamic_ex in day.dynamic_exercises:
                    space = l_ex_str  + 2
                    scheme = ''.join([s.ljust(space) for s in 
                                      self._rendered[week][day][dynamic_ex]['strings']])
                    
                    out_str += '   ' + dynamic_ex.name.ljust(l_ex_name + 3) + scheme + '\n'
                    if verbosity < 2:
                        continue
                    # Advanced information
                    desired_reps = self._rendered[week][day][dynamic_ex]['desired_reps']
                    actual_reps = sum(self._rendered[week][day][dynamic_ex]['reps'])
                    desired_intensity = self._rendered[week][day][dynamic_ex]['desired_intensity']
                    actual_intensity = int(statistics.mean(self._rendered[week][day][dynamic_ex]['intensities']))
                    out_str += '   ' + 'stats'.ljust(l_ex_name+4)
                    out_str += '[reps(actual/desired): {}/{}, intensity(actual/desired): {}/{}]\n'.format(
                            actual_reps, desired_reps, actual_intensity, desired_intensity)
                    
                for static_ex in day.static_exercises:
                    out_str += '   ' + static_ex.name.ljust(l_ex_name + 3) + static_ex.sets_reps + '\n'
                    
                out_str += '\n'                                       
        
        return out_str + '-' * 64
        
    def __str__(self):
        """
        String formatting for readable human output.
        """
        return self.to_text2()

    
    def _autoset_min_reps_consistency(self):
        """
        Sets the program mode to 'weekly', 'daily' or 'exercise'
        by automatically iterating over all exercises.
        """
        
        # Check whether the mode is WEEKLY
        min_reps, max_reps = [], []
        for day in self.days:
            for dynamic_ex in day.dynamic_exercises:
                min_reps.append(dynamic_ex.min_reps)
                max_reps.append(dynamic_ex.max_reps)
        if all_equal(min_reps) and all_equal(max_reps):
            self._min_reps_consistency = 'weekly'
            return None
            
        # Check if mode is DAILY
        for day in self.days:
            min_reps, max_reps = [], []
            for dynamic_ex in day.dynamic_exercises:
                min_reps.append(dynamic_ex.min_reps)
                max_reps.append(dynamic_ex.max_reps)
            if not all_equal(min_reps) or not all_equal(max_reps):
                self._min_reps_consistency = 'exercise'
                return None
        self._min_reps_consistency = 'daily'
        
                

        
        
        
        
        
        
        
        
        


    

if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)

if __name__ == '__main__':
    program = Program('My Program', duration=8,
                      minimum_percentile=0.2,
                      avg_intensity=75,
                      go_to_min=True,
                      reps_per_exercise=25,
                      verbose=False,
                      round_to=2.5,
                      min_reps_consistency='exercise')

    ex = DynamicExercise('Bench press', 130, 150, 2, 8)
    ex2 = DynamicExercise('Squats', 100, 120, 3, 10)
    ex3 = StaticExercise('Biceps')
    day = Day('Day A', exercises=[ex, ex2, ex3])
    program.add_days(day)

    ex = DynamicExercise('Dips', 90, 150, 3, 8,
                         reps = 35,
                         avg_intensity = 75,
                         round_to = None)
    ex5 = StaticExercise('Triceps')
    day = Day(exercises=[ex])
    day.add_exercises(ex5)
    program.add_days(day)

    program.render()
    print(program.to_text2())

    with open('program.html', 'w', encoding='utf-8') as file:
        file.write(program.to_html())




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 15 21:01:51 2017

@author: tommy
"""

from streprogen.utils import (round_to_nearest, prioritized_not_None)
import functools
import warnings

    

class DynamicExercise(object):
    """
    Class for dynamic exercises.
    """
    
    def __init__(self, name, start_weight, end_weight, min_reps = 3, 
                 max_reps = 8, reps = None, avg_intensity = None, 
                 round_to = None):
        """Initialize a new dynamic exercise. A dynamic exercise is rendered by
        the program, and the set/rep scheme will vary from week to week.
    
        Parameters
        ----------
        name
            The name of the exercise, e.g. 'Squats'.
            
        start_weight
            Maximum weight you can lift at the start of the program, e.g. 80.
            
        end_weight
            The goal weight to work towards during the program. This should be
            set in relation to the duration of the training program, e.g. 90.
            
        min_reps
            The minimum number of repetitions for this exercise, e.g. 3.
            
        max_reps
            The maximum number of repetitions for this exercise, e.g. 8.
            
        reps
            The number of baseline repetitions for this exercise. If this
            parameter is set, it will override the global 'reps_per_exercise'
            parameter for the training program. The repetitions will still
            be scaled by the 'reps_scalers' parameter in the training program.
            
        avg_intensity
            The average intensity for this exercise. If set, this will
            override the 'avg_intensity' parameter in the training program. The intensity
            will still be scaled by the 'intensity_scalers' paramter.
            
        round_to
            Round the output to the closest multiple of this number, e.g. 2.5.
            
    
        Returns
        -------
        DynamicExercise
            A DynamicExercise object.
    
    
        Examples
        -------
        >>> bench = DynamicExercise('Bench press', 100, 120, 3, 8)
        """
        self.name = name
        self.start_weight = start_weight
        self.end_weight = end_weight
        self.min_reps = min_reps
        self.max_reps = max_reps
        self.reps = reps
        self.avg_intensity = avg_intensity

        if round_to is None:
            self.round = None
        else:
            self.round = functools.partial(round_to_nearest, nearest = round_to)
        
        if self.start_weight > self.end_weight:
            msg = "Start weight larger than end weight for exericse '{}'."
            warnings.warn(msg.format(self.name))
            
        if self.min_reps > self.max_reps:
            msg = "'min_reps' larger than 'max_reps' for exercise '{}'."
            raise ValueError(msg.format(self.name))
        
    def weekly_growth(self, weeks):
        """Calculate the weekly growth in percentage, and rounds
        to one digit.
    
        Parameters
        ----------
        weeks
            Number of weeks to calculate growth over.

        Returns
        -------
        growth_factor
            A real number such that start * growth_factor** weeks = end.
    
    
        Examples
        -------
        >>> bench = DynamicExercise('Bench press', 100, 120, 3, 8)
        >>> bench.weekly_growth(8)
        2.3
        >>> bench.weekly_growth(4)
        4.7
        """
        start, end = self.start_weight, self.end_weight
        growth_factor = ((end/start)**(1/weeks) - 1)*100
        return round(growth_factor, 1)


        
    def __repr__(self):
        return '{}({})'.format(type(self).__name__, str(self.__dict__)[:60])
        
    def __str__(self):
        """
        Human readable output.
        """

        strvar = ['name', 'start_weight', 'end_weight', 'min_reps', 
                'max_reps', 'reps', 'avg_intensity']
        
        arg_str = ', '.join(['{} = {}'.format(k, self.__dict__[k]) for k 
                             in strvar if self.__dict__[k] is not None])
        
        return '{}({})'.format(type(self).__name__, arg_str)

class StaticExercise(object):
    """
    Class for static exercises.
    """
    
    def __init__(self, name, sets_reps = '4 x 10'):
        """Initialize a new static exercise. A static exercise
        is simply a placeholder for some text.
    
        Parameters
        ----------
        name
            The name of the exercise, e.g. 'Curls'.
            
        sets_reps
            A static set/rep scheme, e.g. '4 x 10', or '10 minutes'.


        Returns
        -------
        StaticExercise
            A StaticExercise object.
    
    
        Examples
        -------
        >>> curls = StaticExercise('Curls', '4 x 10')
        >>> stretching = StaticExercise('Stretching', '10 minutes')
        """
        self.name = name
        self.sets_reps = sets_reps
        
    def __repr__(self):
        return '{}({})'.format(type(self).__name__, str(self.__dict__)[:60])
        
    def __str__(self):
        """
        String formatting for readable human output.
        """

        strvar = ['name', 'sets_reps']
        
        arg_str = ', '.join(['{} = {}'.format(k, self.__dict__[k]) for k 
                             in strvar if self.__dict__[k] is not None])
        
        return '{}({})'.format(type(self).__name__, arg_str)
    
    
    
if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)
    
    bench = DynamicExercise('Bench press', 100, 120, 3, 8)
    print(DynamicExercise('Bench press', 100, 120, 3, 8))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul 17 17:36:22 2017

@author: tommy
"""

from streprogen.exercises import DynamicExercise, StaticExercise

class Day(object):
    """
    A day object is a container for exercises associated with the specific day.
    """
    
    def __init__(self, name = None, exercises = None):
        """Initialize a new day object.
    
        Parameters
        ----------
        name
            The name of the day, e.g. 'Day A'. If no name is given then the day
            will automatically be given a numeric name such as 'Day 1', 'Day 2', etc.
            
        exercises
            A list of exercises. Exercises can also be associated with a day using
            the 'add_exercises' method later on.
    
        Returns
        -------
        Day
            A day object.
    
    
        Examples
        -------
        >>> monday = Day(name = 'Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> monday.add_exercises(curls)
        >>> curls in monday.static_exercises
        True
        """
        self.name = name
        self.dynamic_exercises = []
        self.static_exercises = []
        
        if exercises is not None:
            self.add_exercises(*tuple(exercises))
    
            
    def add_exercises(self, *exercises):
        """Add the exercises to the day. The method will automatically infer
        whether a static or dynamic exercise is passed to it.
    
        Parameters
        ----------
        *exercises
            An unpacked tuple of exercises.
    
    
        Examples
        -------
        >>> monday = Day(name = 'Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> pulldowns = StaticExercise('Pulldowns', '4 x 10')
        >>> monday.add_exercises(curls, pulldowns)
        >>> curls in monday.static_exercises
        True
        >>> pulldowns in monday.static_exercises
        True
        """
        for exercise in list(exercises):
            if isinstance(exercise, DynamicExercise):
                self.dynamic_exercises.append(exercise)
                
            if isinstance(exercise, StaticExercise):
                self.static_exercises.append(exercise)
        
    def __repr__(self):
        return '{}({})'.format(type(self).__name__, str(self.__dict__)[:60])
    
    def __str__(self):
        """
        String formatting for readable human output.
        
        Examples
        -------
        >>> monday = Day(name = 'Monday')
        >>> curls = StaticExercise('Curls', '3 x 12')
        >>> pulldowns = StaticExercise('Pulldowns', '4 x 10')
        >>> monday.add_exercises(curls, pulldowns)
        >>> print(monday)
        Day(name = Monday, static_exercises = [Curls, Pulldowns])
        """
        dyn = ', '.join([d.name for d in self.dynamic_exercises])
        if len(dyn) > 0:
            dyn = "dynamic_exercises = [{}]".format(dyn)
        stat = ', '.join([d.name for d in self.static_exercises])
        if len(stat) > 0:
            stat = "static_exercises = [{}]".format(stat)
        
        out_str = "{}({})".format(type(self).__name__, ', '.join([s for s in 
                      ['name = {}'.format(self.name), dyn, stat]
                      if len(s) > 2]))
        return out_str
    
    
if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)
    


"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))


setup(
    name='streprogen',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.5',

    description='The dynamic strength program generator.',
    long_description= 'Streprogen (short for Strength Program Generator) is a Python package that allows you to write flexible, powerful strength training programs in no time. Streprogen is simple to get started with and has sensible default values, but advanced users can change the underlying algorithms easily too. The training programs scale extremely well, any number of days, exercises and training program durations is possible.',

    # The project's main homepage.
    url='https://github.com/TommyOd/streprogen',

    # Author details
    author='Tommy Odland',
    author_email='',

    # Choose your license
    license='GPLv3',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Healthcare Industry',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='strength training program',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['jinja2', 'pypandoc'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    #package_data={
    #    'sample': ['package_data.dat'],
    #},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #         'sample=sample:main',
    #    ],
    #}
)




# -*- coding: utf-8 -*-
"""
Created on Tue Jul 18 10:47:52 2017

@author: tommy
"""

import random
import functools
import math
import statistics
import operator

class RepellentGenerator(object):
    """
    Generates objects from a domain,
    each time an object is drawn,
    the probability of it being drawn again
    is determined by the probability function.
    """
    
    def __init__(self, domain, probability_func = None, generated = None):
        """Initialize a RepellentGenerator, which is a generator
        where when an object is generated, the probability of it 
        begin generated changes.
    
        Parameters
        ----------
        domain
            A list of objects to generate from, e.g. [1, 2, 3].
        probability_func
            A probability function, it should be decreasing, e.g. lambda x: 1 / 2**x. 
        generated
            A user specified dictionary of the form {element1: num1, element2: num2, ...}
            where num1, num2, ... are the initial states descriping how many times
            the elements element1, element2, ... have been generated. Supplying
            this argument changes the initial probability distribution.
    
    
        Returns
        -------
        RepellentGenerator
            A RepellentGenerator object.
    
    
        Examples
        -------
        >>> domain = [1, 2, 3]
        >>> generator = RepellentGenerator(domain)
        >>> generator.generate_one() in domain
        True
        """
        
        # Initialize the domain
        self.domain = domain
        
        # If no probability function is given, default to exponential
        if probability_func is None:
            self.probability_func = lambda x: 1 / 2**x
        else:
            self.probability_func = probability_func
            
        # If no seed values are given, default to zeros
        if generated is None:
            self.generated = {element:0 for element in domain}
        else:
            self.generated = generated
   
    def generate_one(self):
        """Generate a single element.
    
        Returns
        -------
        element
            An element from the domain.
    
    
        Examples
        -------
        >>> generator = RepellentGenerator(['a', 'b'])
        >>> gen_item = generator.generate_one()
        >>> gen_item in ['a', 'b']
        True
        """
        # Get the weights for all items in the domain
        weights = [self.probability_func(self.generated[element]) 
                    for element in self.domain]
        
        # Sample from the domain using the weights
        element = random.choices(self.domain, weights = weights)[0]
        
        # Update the generated values and return
        self.generated[element] += 1
        
        return element
        
    def yield_from_domain(self, num = 1):
        """Yield 'num' elements from the domain.
    
        Yields
        -------
        A sequence of elements from the domain.
    
    
        Examples
        -------
        >>> domain = ['a', 1]
        >>> generator = RepellentGenerator(domain)
        >>> for element in generator.yield_from_domain(3):
        ...     print(element in domain)
        True
        True
        True
        """
        # Yield the appropriate number of elements
        for i in range(num):
            yield self.generate_one()
            
            
            
def reps_to_intensity(reps, slope = -4.8, constant = 97.5, quadratic = True):
    """A function mapping from repetitions in the range 1 to 12
    to intensities in the range 0 to 100.

    Parameters
    ----------
    reps
        The number of repetitions to map to the intensity range.
    slope
        Slope for the linear function.
    constant
        Constant for the linear function
    quadratic
        If 'True', add a slight quadratic offset.


    Returns
    -------
    intensity
        An intensity value in the range from 0 to 100.


    Examples
    -------
    >>> reps_to_intensity(5, slope = -5, constant = 100, quadratic = False)
    80
    
    >>> reps_to_intensity(8, slope = -5, constant = 100, quadratic = True)
    67.45
    
    >>> reps_to_intensity(8, slope = -5, constant = 100, quadratic = False)
    65
    """
    intensity = constant + slope * (reps - 1)
    if quadratic:
        return intensity + 0.05 * (reps - 1)**2
    else:
        return intensity
    

def progression_linear(week, start_weight, end_weight, start_week, end_week):
    """A linear progression function going through the points
    ('start_week', 'start_weight') and ('end_week', 'end_weight'), evaluated
    in 'week'.

    Parameters
    ----------
    reps
        The number of repetitions to map to the intensity range.
    slope
        Slope for the linear function.
    constant
        Constant for the linear function
    quadratic
        If 'True', add a slight quadratic offset.


    Returns
    -------
    intensity
        An intensity value in the range from 0 to 100.


    Examples
    -------
    >>> progression_linear(week = 2, start_weight = 100, end_weight = 120, 
    ...                    start_week = 1, end_week = 3)
    110.0
    
    >>> progression_linear(3, 100, 140, 1, 5)
    120.0
    """
    # Calculate the slope of the linear function
    slope = (start_weight - end_weight) / (start_week - end_week)
    
    # Return the answer y = slope (x - x_0) + y_0
    return slope * (week - start_week) + start_weight

def progression_sinusoidal(week, start_weight, end_weight, start_week, end_week, 
                           periods = 2, scale = 0.025, offset = 0):
    """
    """
    # Get the linear model
    linear = progression_linear(week, start_weight, end_weight, start_week, end_week)
    
    # Calculate the time period and the argument to the sine function
    time_period = end_week - start_week
    sine_argument = (week - offset - start_week) * (math.pi * 2) / (time_period / periods)
    
    linear_with_sinusoidal = linear * (1 + scale * math.sin(sine_argument))
    return linear_with_sinusoidal



reps_to_intensity_tight = functools.partial(reps_to_intensity, slope = -4)
reps_to_intensity_relaxed = functools.partial(reps_to_intensity, slope = -5.6)


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True)